// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hints

import (
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/middleware"
	"gitlab.bit9.local/psc/ems/hinting/internal/config"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	"github.com/gin-gonic/gin"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"go.uber.org/zap"
)

// Auditor -- Interface to the hint auditing system
type Auditor interface {
	Audit(ctx *gin.Context)
}

// AuditorFactory -- Factory object for the hint auditor
type AuditorFactory func(request *HintRequestMessage, result *HintResult) Auditor

// HintAuditor -- Structure used to manage hint auditing.
type HintAuditor struct {
	request     *HintRequestMessage
	result      *HintResult
	timeDelta   time.Duration
	auditFields []zap.Field
	AuditFailed bool
}

// NewHintAuditor -- Create a new Auditor with a specified time delta
func NewHintAuditor(request *HintRequestMessage, result *HintResult, timeDelta time.Duration) Auditor {
	return &HintAuditor{request: request, result: result, timeDelta: timeDelta}
}

// NewHintAuditorFromCfg -- Create a new Auditor with a time delta as specified by configuration
func NewHintAuditorFromCfg(request *HintRequestMessage, result *HintResult) Auditor {
	return NewHintAuditor(request, result, config.GetInstance().AuditingTimeDelta)
}

// Audit -- Perform an audit
func (auditor *HintAuditor) Audit(ctx *gin.Context) {
	var childSpan opentracing.Span
	if sp := opentracing.SpanFromContext(ctx); sp != nil {
		childSpan = opentracing.StartSpan("HINT AUDIT", opentracing.ChildOf(sp.Context()))
		defer childSpan.Finish()
	}
	auditor.checkHints(childSpan)
	auditor.addTimeDetails()
	if childSpan != nil {
		childSpan.SetTag("device_id", auditor.request.DeviceId)
		childSpan.SetTag("org_key", auditor.request.OrgKey)
		childSpan.SetTag("group_id", auditor.request.DeviceGroupId)
		if auditor.AuditFailed {
			childSpan.SetTag(string(ext.Error), "true")
		}
	}
	if auditor.AuditFailed {
		ctx.Set(string(middleware.AuditLogFields), auditor.auditFields)
	}
}

func (auditor *HintAuditor) checkHints(span opentracing.Span) {
	shouldWarn := !hinttype.CompareHintSets(auditor.request.AuditHintResult.Hints, auditor.result.Hints)
	if shouldWarn {
		auditor.getDetailedtHintDiffs(span)
		auditHints := auditor.request.AuditHintResult.Hints.ToStringArray()
		calculatedHints := auditor.result.Hints.ToStringArray()
		// Consider the audit to have failed only if CBD provided auditing hints because:
		// 1) the microservice owns change times but CBD controls the request last update time;
		//    The microservice time will commonly be after the CBD time.
		//    This disagreement causes spurious extra_hints reports, which will continue
		//    until another update happens.
		// 2) Missing hints are impossible without audit hints
		// 3) extra hints are often the result of product improvements Ex) microservice does not cache updates to change times to ORG and ENV
		// NOTE: the audit results are still logged for investigation
		auditor.AuditFailed = len(auditHints) > 0
		auditor.auditFields = append(auditor.auditFields,
			zap.Strings("audit_hints", auditHints),
			zap.Strings("calculated_hints", calculatedHints),
			zap.Bool("audit_failed", auditor.AuditFailed))

	}
}

func (auditor *HintAuditor) getDetailedtHintDiffs(span opentracing.Span) {
	missingHints := *auditor.request.AuditHintResult.Hints & (^*auditor.result.Hints)
	extraHints := (^*auditor.request.AuditHintResult.Hints) & *auditor.result.Hints
	extraHintsCount := extraHints.Count()
	missingHintsCount := missingHints.Count()
	if span != nil {
		span.SetTag("audit_hints", int(*auditor.request.AuditHintResult.Hints))
		span.SetTag("calculated_hints", int(*auditor.result.Hints))
		span.SetTag("extra_hints", strings.Join(extraHints.ToStringArray(), " "))
		span.SetTag("missing_hints", strings.Join(missingHints.ToStringArray(), " "))
	}
	if extraHintsCount != 0 || missingHintsCount != 0 {
		auditor.auditFields = append(auditor.auditFields,
			zap.Strings("extra_hints", extraHints.ToStringArray()),
			zap.Strings("missing_hints", missingHints.ToStringArray()))
	}
}

func (auditor *HintAuditor) addTimeDetails() {
	auditor.auditFields = append(auditor.auditFields,
		getTimeDetails("audit_last_update_time", "last_update_time",
			auditor.request.AuditHintResult.LastUpdateTime, auditor.result.LastUpdateTime)...)
	auditor.auditFields = append(auditor.auditFields,
		getTimeDetails("audit_last_live_query_time", "last_live_query_time",
			auditor.request.AuditHintResult.LastLiveQueryTime, auditor.result.LastLiveQueryTime)...)
}

func getTimeDetails(labelOne, labelTwo string, one, two *time.Time) []zap.Field {
	switch {
	case one == nil && two == nil:
		return []zap.Field{zap.String(labelOne, "null"), zap.String(labelTwo, "null"), zap.String(labelTwo+"_delta", "-")}
	case one == nil:
		return []zap.Field{zap.String(labelOne, "null"), zap.Time(labelTwo, *two), zap.String(labelTwo+"_delta", "-")}
	case two == nil:
		return []zap.Field{zap.Time(labelOne, *one), zap.String(labelTwo, "null"), zap.String(labelTwo+"_delta", "-")}
	default:
		delta := one.Sub(*two)
		formattedDelta := delta.Round(time.Second).String()
		return []zap.Field{zap.Time(labelOne, *one), zap.Time(labelTwo, *two), zap.String(labelTwo+"_delta", formattedDelta)}
	}
}
