// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hints

import (
	"fmt"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/types"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"
)

type timestampHintAssociation struct {
	changeTime *time.Time
	hintType   hinttype.HintType
}

// HintGenerator -- structure used to manage hint generation
type HintGenerator struct{}

type IHintGenerator interface {
	GenerateHints(request *HintRequest, inAuditMode bool) *HintResult
}

// NewHintGenerator -- Creates a new instance of HintGenerator
func NewHintGenerator() HintGenerator {
	return HintGenerator{}
}

// GenerateHints -- generate hints based on the currently received HintRequest
func (bhg HintGenerator) GenerateHints(request *HintRequest, inAuditMode bool) *HintResult {
	result := NewHintResult()

	bhg.populateGlobalHints(request, result, inAuditMode)
	bhg.populateOrganizationHints(request, result, inAuditMode)
	bhg.populateDeviceGroupHints(request, result, inAuditMode)
	bhg.populateDeviceHints(request, result, inAuditMode)

	logger.BaseLogger().Debug(fmt.Sprintf("Last response time set to %v for device %v",
		result.LastUpdateTime, request.Device))
	return result
}

// populateGlobalHints -- Generate hints based on Env timestamps
//   - as per `apps/deviceServices/src/main/java/com/scargo/app/common/HintGenerator.java`, "populateGlobalHints()"
func (bhg HintGenerator) populateGlobalHints(hRequest *HintRequest, hresult *HintResult, inAuditMode bool) {
	if hRequest.environment == nil {
		return
	}

	var ThreatHunterRulesTimestamp *time.Time = nil
	if hRequest.org != nil {
		ThreatHunterRulesTimestamp = hRequest.org.ChangeTimestamps[orgchangetype.ORG_THREATHUNTER_RULES]
	}
	globalChanges := []timestampHintAssociation{
		{hRequest.environment.CertServiceTime, hinttype.HASH_LIST_REFRESH},
		{hRequest.environment.PolicyServiceUpdateTime, hinttype.NEW_POLICY},
		{ThreatHunterRulesTimestamp, hinttype.NEW_POLICY},
	}
	UpdateHints(hRequest, hresult, globalChanges, inAuditMode)
}

// populateOrganizationHints -- Generate hints based on Org timestamps
//   - as per `apps/deviceServices/src/main/java/com/scargo/app/common/HintGenerator.java`, "populateOrganizationHints()"
//   - NOTE: isAutoUpdateEnabled is not supported any more
func (bhg HintGenerator) populateOrganizationHints(hRequest *HintRequest, hresult *HintResult, inAuditMode bool) {
	if hRequest.org == nil {
		return
	}

	// Handle Universal Change Types
	universalChanges := hRequest.org.ChangeTimestamps
	if universalChanges != nil {
		hresult.processOrgChanges(hRequest, universalChanges, inAuditMode)
	}

	// Handle per-deployment changes
	if hRequest.Device != nil {
		deploymentType := hRequest.Device.DeploymentType
		if deploymentType != nil {
			deploymentTypeChanges := hRequest.org.DeploymentTypeChangeTimestamps[*deploymentType]
			if deploymentTypeChanges != nil {
				hresult.processOrgChanges(hRequest, deploymentTypeChanges, inAuditMode)
			}
		}
	}

	var ransomwareTimestamp *time.Time = nil
	if hRequest.Device != nil {
		deviceType := hRequest.Device.DeviceType
		if deviceType != nil {
			ransomwareTimestamp = hRequest.org.CanaryPayloadTimestamps.GetTimeForDevicetype(*deviceType)
			if ransomwareTimestamp == nil && hRequest.environment != nil {
				ransomwareTimestamp = hRequest.environment.CanaryPayloadTimestamps.GetTimeForDevicetype(*deviceType)
			}
		}
		orgChanges := []timestampHintAssociation{
			{ransomwareTimestamp, hinttype.ZIP_CONTAINER},
		}
		UpdateHints(hRequest, hresult, orgChanges, inAuditMode)
	}
}

// processOrgChanges -- common method for changetype timestamps
//   - as per `apps/deviceServices/src/main/java/com/scargo/app/common/HintGenerator.java`, "populateOrganizationHints()"
//
//goland:noinspection GoMixedReceiverTypes
func (hintResult *HintResult) processOrgChanges(hRequest *HintRequest, changes types.OrgChangeset, inAuditMode bool) {
	orgChanges := []timestampHintAssociation{
		{changes[orgchangetype.ORG_NET_CFG_CHANGE], hinttype.NEW_CONFIG_DATA},
		{changes[orgchangetype.ORG_HASH_LIST_CHANGE], hinttype.HASH_LIST_REFRESH},
		{changes[orgchangetype.ORG_AV_INTEGRATION_CHANGE], hinttype.NEW_POLICY},
		{changes[orgchangetype.ORG_AV_INTEGRATION_CHANGE], hinttype.NEW_CONFIG_DATA},
		{changes[orgchangetype.ORG_UNINSTALL_CODE_CHANGE], hinttype.UNINSTALL_CODE},
		{changes[orgchangetype.ORG_SET_MELTDOWN_REGISTRY_KEY], hinttype.NEW_CONFIG_DATA},
		{changes[orgchangetype.ORG_BLADE_ACTION_CHANGE], hinttype.BLADE_ACTION},
		{changes[orgchangetype.ORG_DISABLE_DEFENSE_RULES], hinttype.NEW_POLICY},
		{changes[orgchangetype.ORG_DISABLE_DEFENSE_RULES], hinttype.NEW_CONFIG_DATA},
		{changes[orgchangetype.ORG_KILL_SWITCH_ON], hinttype.KILL_SWITCH_ON},
		{changes[orgchangetype.ORG_LIVE_QUERY_CHANGE], hinttype.LIVE_QUERY},
	}

	UpdateHints(hRequest, hintResult, orgChanges, inAuditMode)
}

// populateDeviceGroupHints -- Generate hints based on DeviceGroup timestamps
//   - as per `apps/deviceServices/src/main/java/com/scargo/app/common/HintGenerator.java`, "populateDeviceGroupHints()"
func (bhg HintGenerator) populateDeviceGroupHints(hRequest *HintRequest, hintResult *HintResult, inAuditMode bool) {
	if hRequest.deviceGroup == nil {
		return
	}

	var deviceGroupchangeTime *time.Time = nil
	if hRequest.Device != nil {
		deviceGroupchangeTime = hRequest.Device.GroupChangeTime
	}
	deviceGroupChanges := []timestampHintAssociation{
		{hRequest.deviceGroup.LastPolicyChangeTime, hinttype.NEW_POLICY},
		{hRequest.deviceGroup.LastPolicyChangeTime, hinttype.NEW_CONFIG_DATA},
		{deviceGroupchangeTime, hinttype.NEW_POLICY},
		{deviceGroupchangeTime, hinttype.NEW_CONFIG_DATA},
		{hRequest.deviceGroup.LiveQueryChangeTime, hinttype.LIVE_QUERY},
	}
	UpdateHints(hRequest, hintResult, deviceGroupChanges, inAuditMode)
}

// populateDeviceHints -- Generate hints based on DeviceGroup timestamps
//   - as per `apps/deviceServices/src/main/java/com/scargo/app/common/HintGenerator.java`, "populateDeviceHints()"
func (bhg HintGenerator) populateDeviceHints(hRequest *HintRequest, hintResult *HintResult, inAuditMode bool) {
	if hRequest.Device == nil {
		return
	}

	deviceChanges := []timestampHintAssociation{
		{hRequest.Device.SensorConfigUpdateTime, hinttype.NEW_CONFIG_DATA},
		{hRequest.Device.LogsRequestTime, hinttype.CONFER_UPLOAD},
		{hRequest.Device.HashDeleteRequestTime, hinttype.ONETIME_HASH_DELETE_LIST},
		{hRequest.Device.HashUploadRequestTime, hinttype.HASH_LIST_UPLOAD},
		{hRequest.Device.SensorKitUpdateTime, hinttype.SOFTWARE_UPGRADE},
		{hRequest.Device.SensorCertificateUpdateTime, hinttype.SENSOR_CERTIFICATE_UPLOAD},
		{hRequest.Device.PerfmonTriggerTime, hinttype.SENSOR_TRIGGER_PERFMON},
		{hRequest.Device.SensorActionTime, hinttype.SENSOR_ACTION},
		{hRequest.Device.ReRegisterTime, hinttype.REREGISTER},
		{hRequest.Device.LiveQueryChangeTime, hinttype.LIVE_QUERY},
	}

	UpdateHints(hRequest, hintResult, deviceChanges, inAuditMode)
}

// UpdateHints -- Takes a list of hinting checks and goes through them, updating the returned timestamps as
// required and adding the the list of returned hints if any checks pass.
func UpdateHints(hintRequest *HintRequest,
	hintResult *HintResult,
	hintAssociations []timestampHintAssociation,
	inAuditMode bool,
) {
	for _, hintAssociation := range hintAssociations {
		changeTime := hintAssociation.changeTime
		if changeTime == nil { // bypass if check time is nil -- nothing to do
			continue
		}

		// By default, LastUpdateTime is used for checks and responses, except when dealing with LIVE_QUERY
		requestTime := hintRequest.LastUpdateTime
		responseTime := hintResult.LastUpdateTime
		if hintAssociation.hintType == hinttype.LIVE_QUERY {
			requestTime = hintRequest.LastLiveQueryTime
			responseTime = hintResult.LastLiveQueryTime
		}

		// Only do hint checks or timestamps if this is the first time for the device, or if it's update time is older than now
		if requestTime == nil || isTimeBeforeChangeTime(inAuditMode, requestTime, changeTime) {
			if shouldAddHint(hintRequest, hintAssociation.hintType) {
				hintResult.addHint(hintAssociation.hintType)
			}
			if responseTime == nil || isTimeBeforeChangeTime(inAuditMode, responseTime, changeTime) {
				if hintAssociation.hintType == hinttype.LIVE_QUERY {
					hintResult.LastLiveQueryTime = changeTime
				} else {
					hintResult.LastUpdateTime = changeTime
				}
			}
		}
	}
}

func shouldAddHint(hintRequest *HintRequest, hintType hinttype.HintType) bool {
	if hintType == hinttype.KILL_SWITCH_ON {
		if !hintRequest.org.IsKillSwitchOn.Valid || !hintRequest.org.IsKillSwitchOn.Bool {
			return false
		}
	} else if hintType == hinttype.LIVE_QUERY {
		if hintRequest.Version < BASEMESSAGE.LIVE_QUERY_VERSION {
			return false
		}
	}
	return true
}

// isTimeBeforeChangeTime -- Normalize the incoming change time based on operation mode and check for an "older" time.
func isTimeBeforeChangeTime(inAuditMode bool, requestTime, changeTime *time.Time) bool {
	if inAuditMode {
		// For auditing, the timestamp sent to the sensor (which is also sent back to us here as the requestTime) is set by
		// the monolith code and may differ from the timestamp we have previously calculated and stored (the changeTime).
		// Because of this we allow for small differences in those times. Once we are no longer auditing we will be
		// responsible for generating the time sent to the sensor and will not experience this delta. Note that while
		// auditing, the LQ timestamp that we store actually does come from the monolith and we ran into a situation
		// where a LQ timestamp was set twice in less than a second, but we appeared to ignore the 2nd because of this delta check.
		difference := requestTime.Sub(*changeTime)
		return difference < -time.Second
	} else {
		// We truncate here because times coming from the sensor will only have millisecond resolution, but
		// our change times will have microsecond resolution.
		return requestTime.Before(changeTime.Truncate(time.Millisecond))
	}
}
