// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hints

import (
	"database/sql"
	"fmt"
	"log"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

const (
	inAuditMode  = true
	notAuditMode = false
)

// fixed times to make checking easier
var (
	hgUpdate    = time.UnixMilli(1641038400000) // Saturday, January 1, 2022 12:00:00
	hgLivequery = time.UnixMilli(1640952000000) // Friday, December 31, 2021 12:00:00
	hgTime1     = time.UnixMilli(1680609600000) // Tuesday, April 4, 2023 12:00:00
	hgTime2     = time.UnixMilli(1723118400000) // Thursday, August 8, 2024 12:00:00
)

func getTestHintGenerator() HintGenerator {
	return NewHintGenerator()
}

type HintResultMatcher struct {
	hints             []hinttype.HintType
	lastUpdateTime    *time.Time
	lastLiveQueryTime *time.Time
	err               error
}

func NewHintResultMatcher(lastUpdateTime, lastLiveQueryTime *time.Time, hints ...hinttype.HintType) *HintResultMatcher {
	return &HintResultMatcher{lastUpdateTime: lastUpdateTime, lastLiveQueryTime: lastLiveQueryTime, hints: hints}
}

func (hrm *HintResultMatcher) Match(actual interface{}) (success bool, err error) {
	asHintResult := actual.(*HintResult)
	switch {
	case len(hrm.hints) == 0 && asHintResult.AnyHintsSet():
		hrm.err = fmt.Errorf("expected no hints but found: %s", asHintResult.Hints.ToStringArray())
		return false, hrm.err
	case len(hrm.hints) < asHintResult.numHintsSet():
		hrm.err = fmt.Errorf("received additional hints")
		return false, hrm.err
	case !testTimesEqual(hrm.lastLiveQueryTime, asHintResult.LastLiveQueryTime):
		hrm.err = fmt.Errorf("live Query Time didn't match expected: %s found: %s", hrm.lastLiveQueryTime, asHintResult.LastLiveQueryTime)
		return false, hrm.err
	case !testTimesEqual(hrm.lastUpdateTime, asHintResult.LastUpdateTime):
		hrm.err = fmt.Errorf("last Update Time didn't match expected: %s found: %s", hrm.lastUpdateTime, asHintResult.LastUpdateTime)
		return false, hrm.err
	default:
		for _, hint := range hrm.hints {
			if !(asHintResult.IsHintSet(hint)) {
				hrm.err = fmt.Errorf("didn't find expected hint %s", hint.String())
				return false, hrm.err
			}
		}
	}
	return true, nil
}

func (hrm *HintResultMatcher) FailureMessage(actual interface{}) (message string) {
	return fmt.Sprintf("Failed for %v: %v", actual, hrm.err)
}

func (hrm *HintResultMatcher) NegatedFailureMessage(actual interface{}) (message string) {
	return fmt.Sprintf("Negated Failure for %v: %v", actual, hrm.err)
}

// //////////////////////////////////////////////////////////////////////////////
// Begin Hinting tests
// //////////////////////////////////////////////////////////////////////////////

var _ = Describe("HintGenerator", func() {
	generator := getTestHintGenerator()

	It("No change times produces hints set", func() {
		hintRequest := getMockHintRequest(nil, nil)
		hintResult := generator.GenerateHints(hintRequest, inAuditMode)
		Expect(hintResult).To(NewHintResultMatcher(nil, nil))
	})

	// ----- Verify that missing request parameters are handled without error

	Context("nil structure checks", func() {
		It("env is nil", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.environment = nil
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
		It("org is nil", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.org = nil
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
		It("group is nil", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.deviceGroup = nil
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
		It("device is nil", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device = nil
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
	})

	// ----- Environment hints --------------------

	Context("Global Hint Tests", func() {
		Context("CertServiceTime", func() {
			It("with REQ_HELLO we get hint", func() {
				hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
				hintRequest.environment.CertServiceTime = &hgTime1
				hintResult := generator.GenerateHints(hintRequest, inAuditMode)
				Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.HASH_LIST_REFRESH))
			})
		})

		Context("PolicyServiceUpdateTime", func() {
			It("with REQ_HELLO we get hint", func() {
				hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
				hintRequest.environment.PolicyServiceUpdateTime = &hgTime1
				hintResult := generator.GenerateHints(hintRequest, inAuditMode)
				Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.NEW_POLICY))
			})
		})

		Context("ThreatHunterRulesTimestamp ", func() {
			// NOTE: these tests depend on org being defined and ChangeTimestamps[ORG_THREATHUNTER_RULES] being defined
			It("with REQ_HELLO we get hint", func() {
				hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
				hintRequest.org.ChangeTimestamps[orgchangetype.ORG_THREATHUNTER_RULES] = &hgTime1
				hintResult := generator.GenerateHints(hintRequest, inAuditMode)
				Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.NEW_POLICY))
			})
			It("No hint when org's ChangeTimestamps has no ORG_THREATHUNTER_RULES defined", func() {
				// NOTE: Same as if org is not defined
				hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
				hintRequest.org.ChangeTimestamps[orgchangetype.ORG_THREATHUNTER_RULES] = nil
				hintResult := generator.GenerateHints(hintRequest, inAuditMode)
				Expect(hintResult).To(NewHintResultMatcher(nil, nil))
			})
		})
	})

	// ----- Organization tests --------------------

	// This data is used for org change type where a specific messagecode is NOT specified -- hints expected
	testData := []struct {
		changetype orgchangetype.OrgChangeType
		hints      []hinttype.HintType
	}{
		{changetype: orgchangetype.ORG_THREATHUNTER_RULES, hints: []hinttype.HintType{hinttype.NEW_POLICY}},
		{changetype: orgchangetype.ORG_NET_CFG_CHANGE, hints: []hinttype.HintType{hinttype.NEW_CONFIG_DATA}},
		{changetype: orgchangetype.ORG_HASH_LIST_CHANGE, hints: []hinttype.HintType{hinttype.HASH_LIST_REFRESH}},
		{changetype: orgchangetype.ORG_AV_INTEGRATION_CHANGE, hints: []hinttype.HintType{hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA}},
		{changetype: orgchangetype.ORG_KILL_SWITCH_ON, hints: []hinttype.HintType{hinttype.KILL_SWITCH_ON}},
		{changetype: orgchangetype.ORG_UNINSTALL_CODE_CHANGE, hints: []hinttype.HintType{hinttype.UNINSTALL_CODE}},
		{changetype: orgchangetype.ORG_SET_MELTDOWN_REGISTRY_KEY, hints: []hinttype.HintType{hinttype.NEW_CONFIG_DATA}},
		{changetype: orgchangetype.ORG_LIVE_QUERY_CHANGE, hints: []hinttype.HintType{hinttype.LIVE_QUERY}},
		{changetype: orgchangetype.ORG_BLADE_ACTION_CHANGE, hints: []hinttype.HintType{hinttype.BLADE_ACTION}},
		{changetype: orgchangetype.ORG_DISABLE_DEFENSE_RULES, hints: []hinttype.HintType{hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA}},
	}

	Context("Org ChangeTimestamps with REQ_HELLO", func() {
		for i := range testData {
			testParams := testData[i]
			It(fmt.Sprintf("Test %s -> %v", testParams.changetype.String(), testParams.hints), func() {
				log.Printf("Test %s -> %v", testParams.changetype.String(), testParams.hints)
				hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
				hintRequest.org.ChangeTimestamps[testParams.changetype] = &hgTime1
				if testParams.changetype == orgchangetype.ORG_KILL_SWITCH_ON {
					// NOTE: Special case: ORG_KILL_SWITCH_ON
					hintRequest.org.IsKillSwitchOn = sql.NullBool{Valid: true, Bool: true}
				}

				hintResult := generator.GenerateHints(hintRequest, inAuditMode)
				if testParams.changetype == orgchangetype.ORG_LIVE_QUERY_CHANGE {
					// NOTE: Special case: ORG_LIVE_QUERY_CHANGE
					Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1, testParams.hints...))
				} else {
					Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, testParams.hints...))
				}
			})
		}
	})

	Context("Special Case hint generation", func() {
		// NOTE: KILL_SWITCH_ON never sent if value is false
		It("Test ORG_KILL_SWITCH_ON & false value -> []", func() {
			log.Printf("Test ORG_KILL_SWITCH_ON & false value -> []")
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_KILL_SWITCH_ON] = &hgTime1
			hintRequest.org.IsKillSwitchOn = sql.NullBool{Valid: true, Bool: false}

			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil))
		})

		// NOTE: LIVE_QUERY never sent if value is false
		It("Test ORG_LIVE_QUERY_CHANGE & Version too old -> []", func() {
			log.Printf("Test ORG_LIVE_QUERY_CHANGE & Version too old -> []")
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_LIVE_QUERY_CHANGE] = &hgTime1
			hintRequest.Version = 1

			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1))
		})
	})

	// validate hinting for all deployment types
	Context("Org ChangeTimestamps involving DeploymentTypeChangeTimestamps", func() {
		for i := range deploymenttype.AllDeploymentTypes {
			deploymentType := deploymenttype.AllDeploymentTypes[i]
			Context(fmt.Sprintf("Validating DeploymentTypeChangeTimestamps DeploymentType %v", deploymentType), func() {
				for j := range testData {
					testParams := testData[j]
					// NOTE: No hints for ORG_THREATHUNTER_RULES within DeploymentTypeChangeTimestamps
					if testParams.changetype == orgchangetype.ORG_THREATHUNTER_RULES {
						continue
					}

					It(fmt.Sprintf("Test device %s on %v -> %v", testParams.changetype.String(), deploymentType, testParams.hints), func() {
						log.Printf("Test device %s on %v -> %v", testParams.changetype.String(), deploymentType, testParams.hints)
						hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
						hintRequest.Device.DeploymentType = &deploymentType
						hintRequest.org.GetOrCreateDeploymentTypeChanges(deploymentType)[testParams.changetype] = &hgTime2
						if testParams.changetype == orgchangetype.ORG_KILL_SWITCH_ON {
							// NOTE: Special case: ORG_KILL_SWITCH_ON
							hintRequest.org.IsKillSwitchOn = sql.NullBool{Valid: true, Bool: true}
						}

						hintResult := generator.GenerateHints(hintRequest, inAuditMode)
						if testParams.changetype == orgchangetype.ORG_LIVE_QUERY_CHANGE {
							// NOTE: Special case: ORG_LIVE_QUERY_CHANGE
							Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime2, testParams.hints...))
						} else {
							Expect(hintResult).To(NewHintResultMatcher(&hgTime2, nil, testParams.hints...))
						}
					})

					It(fmt.Sprintf("Test unmatched device %s on %v -> []", testParams.changetype.String(), deploymentType), func() {
						log.Printf("Test unmatched device %s on %v -> []", testParams.changetype.String(), deploymentType)
						hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
						mismatch := deploymenttype.AllDeploymentTypes[(i+1)%len(deploymenttype.AllDeploymentTypes)]
						hintRequest.Device.DeploymentType = &mismatch
						hintRequest.org.GetOrCreateDeploymentTypeChanges(deploymentType)[testParams.changetype] = &hgTime1
						hintRequest.org.GetOrCreateDeploymentTypeChanges(mismatch)[testParams.changetype] = nil // make sure the mismatch DeploymentType is empty
						hintResult := generator.GenerateHints(hintRequest, inAuditMode)
						Expect(hintResult).To(NewHintResultMatcher(nil, nil))
					})

					It(fmt.Sprintf("Test device %s, %v not set -> []", testParams.changetype.String(), deploymentType), func() {
						log.Printf("Test device %s, %v not set -> []", testParams.changetype.String(), deploymentType)
						hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
						hintRequest.org.GetOrCreateDeploymentTypeChanges(deploymentType)[testParams.changetype] = nil
						hintResult := generator.GenerateHints(hintRequest, inAuditMode)
						Expect(hintResult).To(NewHintResultMatcher(nil, nil))
					})
				}
			})
		}
	})

	Context("Org ChangeTimestamps involving CanaryPayloadTimestamps", func() {
		It("Environment CanaryPayloadTimestamps with REQ_HELLO = hint", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.environment.CanaryPayloadTimestamps[*hintRequest.Device.DeviceType] = hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.ZIP_CONTAINER))
		})
		It("Org CanaryPayloadTimestamps with REQ_HELLO = hint", func() {
			hintRequest := getMockHintRequest(&hgUpdate, nil)
			hintRequest.org.CanaryPayloadTimestamps[*hintRequest.Device.DeviceType] = hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.ZIP_CONTAINER))
		})
		It("deviceType is nil = no hint", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.org.CanaryPayloadTimestamps[*hintRequest.Device.DeviceType] = hgTime1
			hintRequest.Device.DeviceType = nil

			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
		It("unknown deviceType = no hint", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.org.CanaryPayloadTimestamps[*hintRequest.Device.DeviceType] = hgTime1
			deviceType := devicetype.FromString("BOGUS_DEVICE")
			hintRequest.Device.DeviceType = &deviceType
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})
	})

	// ----- Begin Device Group tests --------------------

	Context("Group Hints", func() {
		It("Send Hints for Device's Group GroupChangeTime Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.GroupChangeTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA))
		})

		It("Send Hints for Device Group's LastPolicyChangeTime Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.deviceGroup.LastPolicyChangeTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA))
		})

		It("Send Hints for Device Group's LiveQuery Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.deviceGroup.LiveQueryChangeTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1, hinttype.LIVE_QUERY))
		})

		It("Do not send Hints when backend time is behind ", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgTime1)
			hintRequest.deviceGroup.LiveQueryChangeTime = &hgLivequery
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, nil))
		})

		It("Update time but dont send hint if the protocol version is older", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.deviceGroup.LiveQueryChangeTime = &hgTime1
			hintRequest.Version = 0
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1))
		})
	})

	// ----- Begin Device tests --------------------

	Context("Device Hints", func() {
		It("Send Hints for SensorConfigUpdateTime", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorConfigUpdateTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.NEW_CONFIG_DATA))
		})

		It("Send Hints for Logs Request Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.LogsRequestTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.CONFER_UPLOAD))
		})

		It("Send Hints for Hash Delete Request Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.HashDeleteRequestTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.ONETIME_HASH_DELETE_LIST))
		})

		It("Send Hints for Hash Upload Request Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.HashUploadRequestTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.HASH_LIST_UPLOAD))
		})

		It("Send Hints for Sensor Kit Update Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorKitUpdateTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.SOFTWARE_UPGRADE))
		})

		It("Send Hints for Sensor Certificate Update Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorCertificateUpdateTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.SENSOR_CERTIFICATE_UPLOAD))
		})

		It("Send Hints for Perfmon Trigger Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.PerfmonTriggerTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.SENSOR_TRIGGER_PERFMON))
		})

		It("Send Hints for Sensor Action Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorActionTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.SENSOR_ACTION))
		})

		It("Send Hints for Re-register Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.ReRegisterTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil, hinttype.REREGISTER))
		})

		It("Send Hints for Live Query Change Timestamp", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.LiveQueryChangeTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1, hinttype.LIVE_QUERY))
		})
		It("Send no Hints for Live Query Change Timestamp, when version is old", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.LiveQueryChangeTime = &hgTime1
			hintRequest.Version = 1 // current version is 17+
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(nil, &hgTime1))
		})
	})

	Context("Multiple Global Hints", func() {
		It("Global cert service time, PolicyService Update time and ThreatHunter rules time is changed", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.environment.CertServiceTime = &hgTime1
			hintRequest.environment.PolicyServiceUpdateTime = &hgTime1
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_THREATHUNTER_RULES] = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.HASH_LIST_REFRESH,
				hinttype.NEW_POLICY))
		})
	})

	Context("Multiple organization Hints", func() {
		deploymentType := deploymenttype.ENDPOINT
		It("Hash list, Net config is changed and defense rules are disabled", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.DeploymentType = &deploymentType
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_NET_CFG_CHANGE] = &hgTime1
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_HASH_LIST_CHANGE] = &hgTime1
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_DISABLE_DEFENSE_RULES] = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.NEW_CONFIG_DATA,
				hinttype.HASH_LIST_REFRESH,
				hinttype.NEW_POLICY))
		})
	})

	Context("Multiple Group Hints", func() {
		It("Device's Group time, last update time is changed", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.GroupChangeTime = &hgTime1
			hintRequest.deviceGroup.LastPolicyChangeTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.NEW_POLICY,
				hinttype.NEW_CONFIG_DATA))
		})
	})

	Context("Multiple Device Hints", func() {
		It("SensorConfigUpdateTime, HashDeleteRequestTime, HashUploadRequestTime, PerfmonTriggerTime are changed", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorConfigUpdateTime = &hgTime1
			hintRequest.Device.HashDeleteRequestTime = &hgTime1
			hintRequest.Device.HashUploadRequestTime = &hgTime1
			hintRequest.Device.PerfmonTriggerTime = &hgTime1
			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.SENSOR_TRIGGER_PERFMON,
				hinttype.ONETIME_HASH_DELETE_LIST,
				hinttype.HASH_LIST_UPLOAD,
				hinttype.NEW_CONFIG_DATA))
		})
	})

	Context("Random Hints", func() {
		It("Global: PolicyService, org: Hash list, Update time, Group: Device's Group time,"+
			" Device: SensorConfigUpdateTime are changed", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorConfigUpdateTime = &hgTime1
			hintRequest.Device.GroupChangeTime = &hgTime1
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_HASH_LIST_CHANGE] = &hgTime1
			hintRequest.environment.PolicyServiceUpdateTime = &hgTime1

			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.NEW_CONFIG_DATA,
				hinttype.HASH_LIST_REFRESH,
				hinttype.NEW_POLICY,
				hinttype.NEW_CONFIG_DATA))
		})

		It("Global: PolicyService, org: Hash list, Update time, Group: Device's Group time,"+
			" Device: SensorConfigUpdateTime are changed (different audit mode)", func() {
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorConfigUpdateTime = &hgTime1
			hintRequest.Device.GroupChangeTime = &hgTime1
			hintRequest.org.ChangeTimestamps[orgchangetype.ORG_HASH_LIST_CHANGE] = &hgTime1
			hintRequest.environment.PolicyServiceUpdateTime = &hgTime1

			hintResult := generator.GenerateHints(hintRequest, notAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1,
				nil,
				hinttype.NEW_CONFIG_DATA,
				hinttype.HASH_LIST_REFRESH,
				hinttype.NEW_POLICY,
				hinttype.NEW_CONFIG_DATA))
		})

		It("Validate DSER-48304 situation", func() {
			// As part of Sensor Gateway actions we create data in Sensor Action (subType = CHANGE_BACKEND_SERVER) with
			// a status=TRIGGERED and we update the Device Table timestamps accordingly.(DEV_SENSOR_CONFIG_UPDATE).
			//
			// We expect that upon next check-in from the Sensor a hint to be sent to the sensor so that it can force
			// pull the configurations. We are noticing that at times the Hints are delayed OR no hints are sent.
			hintRequest := getMockHintRequest(&hgUpdate, &hgLivequery)
			hintRequest.Device.SensorConfigUpdateTime = &hgTime1

			hintResult := generator.GenerateHints(hintRequest, inAuditMode)
			Expect(hintResult).To(NewHintResultMatcher(&hgTime1, nil,
				hinttype.NEW_CONFIG_DATA))
		})
	})
})

func TestHintGenerator(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Hint Generator Suite")
}

func BenchmarkHintGenerator(b *testing.B) {
	generator := getTestHintGenerator()
	hintRequest := getWorstCaseHintRequest(&hgUpdate, &hgUpdate)
	for n := 0; n < b.N; n++ {
		generator.GenerateHints(hintRequest, inAuditMode)
	}
}

func testTimesEqual(expected, found *time.Time) bool {
	timeOk := false
	if expected == nil {
		if found == nil {
			timeOk = true
		} else {
			timeOk = false
		}
	} else {
		if found == nil {
			timeOk = false
		} else {
			timeOk = found.Equal(*expected)
		}
	}
	return timeOk
}
